@@ -1,5 +1,5 @@
 __author__ = "desultory"
-__version__ = "3.2.4"
+__version__ = "3.7.3"
 
 from pathlib import Path
 
@@ -20,7 +20,9 @@ CRYPTSETUP_PARAMETERS = [
     *CRYPTSETUP_KEY_PARAMETERS,
     "try_nokey",
     "include_key",
+    "include_header",
     "validate_key",
+    "validate_header",
 ]
 
 
@@ -54,32 +56,19 @@ def _process_cryptsetup_key_types_multi(self, key_type: str, config: dict) -> No
 
 
 @contains("validate", "Skipping cryptsetup keyfile validation.", log_level=30)
-def _validate_crypysetup_key(self, key_paramters: dict) -> None:
+def _validate_crypysetup_key(self, key_parameters: dict) -> None:
     """Validates the cryptsetup key"""
-    if key_paramters.get("include_key"):
+    if key_parameters.get("include_key"):
         return self.logger.info("Skipping key validation for included key.")
-    elif key_paramters.get("validate_key") is False:
-        return self.logger.info("Skipping key validation for: %s" % key_paramters["key_file"])
+    elif key_parameters.get("validate_key") is False:
+        return self.logger.info("Skipping key validation for: %s" % key_parameters["key_file"])
 
-    key_path = Path(key_paramters["key_file"])
+    key_path = Path(key_parameters["key_file"])
 
-    if not key_path.is_file():
-        if self["cryptsetup_keyfile_validation"]:
+    if self["cryptsetup_keyfile_validation"] and key_parameters.get("validate_key") is not False:
+        if not key_path.exists():  # Do a preliminary check to see if the key file exists
             raise FileNotFoundError("Key file not found: %s" % key_path)
-        else:
-            return self.logger.error("Key file not found: %s" % key_path)
-
-    key_copy = key_path
-    while parent := key_copy.parent:
-        if parent == Path("/"):
-            if self["cryptsetup_keyfile_validation"]:
-                raise ValueError("No mount is defined for external key file: %s" % key_path)
-            else:
-                return self.logger.critical("No mount is defined for external key file: %s" % key_path)
-        if str(parent).lstrip("/") in self["mounts"]:
-            self.logger.debug("Found mount for key file: %s" % parent)
-            break
-        key_copy = parent
+        self["check_included_or_mounted"] = key_path  # Add a proper check for later
 
 
 @contains("validate", "Skipping cryptsetup configuration validation.", log_level=30)
@@ -134,9 +123,10 @@ def _process_cryptsetup_multi(self, mapped_name: str, config: dict) -> None:
             if value := self["cryptsetup_key_types"][key_type].get(parameter):
                 config[parameter] = value.format(**config)
 
-    # Include the key file if include_key is set
-    if config.get("include_key"):
-        self["dependencies"] = config["key_file"]
+    # Include the key file or header if set
+    for file_type in ["key", "header"]:
+        if config.get(f"include_{file_type}"):
+            self["dependencies"] = config[f"{file_type}_file"]
 
     if not config.get("retries"):
         self.logger.info(
@@ -147,17 +137,129 @@ def _process_cryptsetup_multi(self, mapped_name: str, config: dict) -> None:
     self["cryptsetup"][mapped_name] = config
 
 
+def _get_dm_info(self, mapped_name: str) -> dict:
+    """Gets the device mapper information for a particular device."""
+    for device_info in self["_vblk_info"].values():
+        if device_info["name"] == mapped_name:
+            return device_info
+    raise KeyError("No device mapper information found for: %s" % mapped_name)
+
+
+def _get_dm_slave_info(self, device_info: dict) -> (str, dict):
+    """Gets the device mapper slave information for a particular device."""
+    slave_source = device_info["slaves"][0]
+    try:
+        slave_device = f"/dev/{slave_source}"
+        return slave_device, self["_blkid_info"][slave_device]
+    except KeyError:
+        slave_device = f"/dev/mapper/{slave_source}"
+        return slave_device, self["_blkid_info"][slave_device]
+
+
+def _read_cryptsetup_header(self, mapped_name: str, slave_device: str = None) -> dict:
+    """Reads LUKS header information from a device or header file into a dict"""
+    from json import loads
+
+    header_file = self["cryptsetup"][mapped_name].get("header_file")
+    if not header_file:
+        if slave_device:
+            header_file = slave_device
+        else:
+            slave_device, _ = _get_dm_slave_info(self, _get_dm_info(self, mapped_name))
+            header_file = slave_device
+    try:  # Try to read the header, return data, decoded and loaded, as a dictionary
+        luks_info = loads(
+            self._run(
+                ["cryptsetup", "luksDump", "--dump-json-metadata", header_file], fail_silent=True, fail_hard=True
+            ).stdout.decode()
+        )
+        self.logger.debug("[%s] LUKS header information: %s" % (mapped_name, luks_info))
+        raw_luks_info = self._run(["cryptsetup", "luksDump", header_file]).stdout.decode().split("\n")
+        # --dump-json-metadata does not include the UUID, so we need to parse it from the raw output
+        for line in raw_luks_info:
+            if "UUID" in line:
+                luks_info["uuid"] = line.split()[1]
+                break
+        return luks_info
+    except RuntimeError as e:
+        if not self["cryptsetup"][mapped_name].get("header_file"):
+            self.logger.error("Unable to read LUKS header: %s" % e)
+        else:
+            self.logger.warning("Cannot read detached LUKS header for validation: %s" % e)
+    return {}
+
+
+def _detect_luks_aes_module(self, luks_cipher_name: str) -> None:
+    """Using the cipher name from the LUKS header, detects the corresponding kernel module."""
+    aes_type = luks_cipher_name.split("-")[1]  # Get the cipher type from the name
+    self["_kmod_auto"] = aes_type  # Add the aes type to the kernel modules
+
+    crypto_name = f"{aes_type}(aes)"  # Format the name like the /proc/crypto entry
+    crypto_config = self["_crypto_ciphers"][crypto_name]
+    if crypto_config["module"] == "kernel":
+        self.logger.debug("Cipher kernel modules are builtin: %s" % crypto_name)
+    else:
+        self.logger.info("[%s] Adding kernel module for LUKS cipher: %s" % (crypto_name, crypto_config["module"]))
+        self["_kmod_auto"] = crypto_config["module"]
+
+
+def _detect_luks_header_aes(self, luks_info: dict) -> dict:
+    """Checks the cipher type in the LUKS header, reads /proc/crypto to find the
+    corresponding driver. If it's not builtin, adds the module to the kernel modules."""
+    for keyslot in luks_info.get("keyslots", {}).values():
+        if keyslot.get("area", {}).get("encryption").startswith("aes"):
+            _detect_luks_aes_module(self, keyslot["area"]["encryption"])
+    for segment in luks_info.get("segments", {}).values():
+        if segment.get("encryption").startswith("aes"):
+            _detect_luks_aes_module(self, segment["encryption"])
+
+
+def _detect_luks_header_sha(self, luks_info: dict) -> dict:
+    """Reads the hash algorithm from the LUKS header,
+    enables the corresponding kernel module using _crypto_ciphers"""
+    for keyslot in luks_info.get("keyslots", {}).values():
+        if keyslot.get("af", {}).get("hash").startswith("sha"):
+            self["kernel_modules"] = self._crypto_ciphers[keyslot["af"]["hash"]]["driver"]
+    for digest in luks_info.get("digests", {}).values():
+        if digest.get("hash").startswith("sha"):
+            self["kernel_modules"] = self._crypto_ciphers[digest["hash"]]["driver"]
+
+
+@contains("cryptsetup_header_validation", "Skipping cryptsetup header validation.", log_level=30)
+def _validate_cryptsetup_header(self, mapped_name: str) -> None:
+    """Validates configured cryptsetup volumes against the LUKS header."""
+    cryptsetup_info = self["cryptsetup"][mapped_name]
+    if cryptsetup_info.get("validate_header") is False:
+        return self.logger.warning("Skipping cryptsetup header validation for: %s" % mapped_name)
+
+    luks_info = _read_cryptsetup_header(self, mapped_name)
+    if not luks_info:
+        raise ValueError("[%s] Unable to read LUKS header." % mapped_name)
+
+    if uuid := cryptsetup_info.get("uuid"):
+        if luks_info.get("uuid") != uuid:
+            raise ValueError(
+                "[%s] LUKS UUID mismatch, found '%s', expected: %s" % (mapped_name, luks_info["uuid"], uuid)
+            )
+
+    _detect_luks_header_aes(self, luks_info)
+    _detect_luks_header_sha(self, luks_info)
+
+    if not self["argon2"]:  # if argon support was not detected, check if the header wants it
+        for keyslot in luks_info.get("keyslots", {}).values():
+            if keyslot.get("kdf", {}).get("type") == "argon2id":
+                raise FileNotFoundError("[%s] Missing cryptsetup dependency: libargon2.so" % mapped_name)
+
+    if 'header_file' in cryptsetup_info:
+        self["check_included_or_mounted"] = cryptsetup_info["header_file"]  # Add the header file to the check list
+
+
 @contains("hostonly", "Skipping cryptsetup device check.", log_level=30)
 def _validate_cryptsetup_device(self, mapped_name) -> None:
     """Validates a cryptsetup device against the device mapper information,
     blkid information, and cryptsetup information.
     Uses `cryptsetup luksDump` to check that the device is a LUKS device."""
-    for device_info in self["_vblk_info"].values():
-        if device_info["name"] == mapped_name:
-            dm_info = device_info  # Get the device mapper information
-            break
-    else:
-        raise ValueError("No device mapper information found for: %s" % mapped_name)
+    dm_info = _get_dm_info(self, mapped_name)
 
     if not dm_info["uuid"].startswith("CRYPT-LUKS"):  # Ensure the device is a crypt device
         raise ValueError("Device is not a crypt device: %s" % dm_info)
@@ -165,16 +267,10 @@ def _validate_cryptsetup_device(self, mapped_name) -> None:
     cryptsetup_info = self["cryptsetup"][mapped_name]  # Get the cryptsetup information
     if cryptsetup_info.get("validate") is False:
         return self.logger.warning("Skipping cryptsetup device validation: %s" % mapped_name)
-    slave_source = dm_info["slaves"][0]  # Get the slave source
 
-    try:  # Get the blkid information
-        slave_device = f"/dev/{slave_source}"
-        blkid_info = self["_blkid_info"][slave_device]
-    except KeyError:
-        slave_device = f"/dev/mapper/{slave_source}"
-        blkid_info = self["_blkid_info"][slave_device]
+    slave_device, blkid_info = _get_dm_slave_info(self, dm_info)  # Get the blkid information
 
-    for token_type in ["partuuid", "uuid"]:  # Validate the uuid/partuuid token
+    for token_type in ["partuuid", "uuid"]:  # Validate the uuid/partuuid token against blkid info
         if cryptsetup_token := cryptsetup_info.get(token_type):
             if blkid_info.get(token_type) != cryptsetup_token:
                 raise ValueError(
@@ -185,40 +281,7 @@ def _validate_cryptsetup_device(self, mapped_name) -> None:
     else:
         raise ValueError("[%s] No UUID or PARTUUID set for LUKS source: %s" % (mapped_name, cryptsetup_info))
 
-    try:
-        # Get the luks header info, remove tabs, and split the output
-        luks_header_source = cryptsetup_info.get("header_file") or slave_device
-        luks_info = (
-            self._run(["cryptsetup", "luksDump", luks_header_source]).stdout.decode().replace("\t", "").split("\n")
-        )  # Check that the device is a LUKS device
-        self.logger.debug("[%s] LUKS information:\n%s" % (mapped_name, luks_info))
-    except RuntimeError as e:
-        luks_info = []
-        if not cryptsetup_info.get("header_file"):
-            return self.logger.error("[%s] Unable to read LUKS header: %s" % (mapped_name, e))
-        self.logger.warning("[%s] Cannot read detached LUKS header for validation: %s" % (mapped_name, e))
-
-    if token_type == "uuid":  # Validate the LUKS UUID using the header
-        for line in luks_info:
-            if "UUID" in line:
-                if line.split()[1] != cryptsetup_token:
-                    raise ValueError(
-                        "[%s] LUKS UUID mismatch, found '%s', expected: %s"
-                        % (mapped_name, line.split()[1], cryptsetup_token)
-                    )
-                break
-        else:
-            raise ValueError("[%s] Unable to validate LUKS UUID: %s" % (mapped_name, cryptsetup_token))
-
-    if "Cipher:     aes-xts-plain64" in luks_info:
-        self["kernel_modules"] = "crypto_xts"
-
-    if not self["argon2"]:
-        if cryptsetup_info.get("header_file"):  # A header may be specified but unavailable
-            self.logger.error("[%s] Unable to check: libargon2.so" % mapped_name)
-        if "PBKDF:      argon2id" in luks_info:  # If luks info is found, and argon is used, raise an error
-            raise FileNotFoundError("[%s] Missing cryptsetup dependency: libargon2.so" % mapped_name)
-        self.logger.error("[%s] Unable to validate argon support for LUKS: %s" % (mapped_name, luks_info))
+    _validate_cryptsetup_header(self, mapped_name)  # Run header validation, mostly for crypto modules
 
 
 def detect_argon2(self) -> None:
@@ -236,6 +299,27 @@ def detect_argon2(self) -> None:
     self["argon2"] = argon
 
 
+@contains("hostonly")
+def detect_ciphers(self) -> None:
+    """Populates _crypto_ciphers using /proc/crypto"""
+
+    def get_value(line):
+        return line.split(":")[1].strip()
+
+    with open("/proc/crypto") as crypto_file:
+        current_name = None
+        for line in crypto_file:
+            if line.startswith("name"):
+                current_name = get_value(line)
+                self._crypto_ciphers[current_name] = {}
+            elif not current_name:
+                continue  # Skip lines until a name is found
+            elif line.startswith("driver"):
+                self._crypto_ciphers[current_name]["driver"] = get_value(line)
+            elif line.startswith("module"):
+                self._crypto_ciphers[current_name]["module"] = get_value(line)
+
+
 @contains("validate", "Skipping cryptsetup configuration validation.", log_level=30)
 def _validate_luks_config(self, mapped_name: str) -> None:
     """Checks that a LUKS config portion is valid."""
@@ -311,8 +395,7 @@ def open_crypt_device(self, name: str, parameters: dict) -> list[str]:
         "    fi",
     ]
 
-
-    reset_command = parameters.get("reset_command", 'continue')
+    reset_command = parameters.get("reset_command", "continue")
     # When there is a key command, evaluate it into $key_data
     if "key_command" in parameters:
         self.logger.debug("[%s] Using key command: %s" % (name, parameters["key_command"]))
