@@ -1,5 +1,5 @@
 __author__ = "desultory"
-__version__ = "5.6.0"
+__version__ = "5.6.2"
 
 from pathlib import Path
 from typing import Union
@@ -107,7 +107,7 @@ def _process_mount(self, mount_name: str, mount_config, mount_class="mounts") ->
             if "ugrd.fs.bcachefs" not in self["modules"]:
                 self.logger.info("Auto-enabling module: bcachefs")
                 self["modules"] = "ugrd.fs.bcachefs"
-        elif mount_type not in ["proc", "sysfs", "devtmpfs", "tmpfs", "devpts"]:
+        elif mount_type not in ["proc", "sysfs", "devtmpfs", "squashfs", "tmpfs", "devpts"]:
             self.logger.warning("Unknown mount type: %s" % mount_type)
 
     self[mount_class][mount_name] = mount_config
@@ -428,6 +428,9 @@ def _autodetect_dm(self, mountpoint, device=None) -> None:
     elif blkid_info.get("type") == "linux_raid_member":
         autodetect_raid(self, source_device, dev_name, blkid_info)
     else:
+        if "type" not in blkid_info:
+            self.logger.error("If LUKS headers are detached, they must be configured with the corresponding mapped device name.")
+            raise RuntimeError("[%s] No type found for device mapper device: %s" % (dev_name, source_device))
         raise RuntimeError("Unknown device mapper device type: %s" % blkid_info.get("type"))
 
     autodetect_mount_kmods(self, slave_source)
